/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1996 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"
#include "wantgui.h"

#ifdef UNIX
#define NAME_START	39	/* column file name starts in */
#define AFTER_DIR	"/"	/* char after name of directory */
#define SIZE_COL	16	/* column with size of file */
#define DATESTART	26	/* column with file date */
#define ISDIR_START	2	/* column to check for directory in */
#define DIRMARK		"d"	/* what must be in column, if directory */
#define DIRSEARCH	"^..d"	/* search for this regex to find dirs */
#else
#define DIRMARK		"<DIR>"
#define DIRSEARCH	" %<DIR%> " /* search for this regex to find dirs */
#if 1
#define NAME_START	33
#define ISDIR_START	23
#define AFTER_DIR	"\n"
#define SIZE_COL	18	/* column with size of file */
#define DATESTART	0	/* column with file date */
#else
#define NAME_START	2
#define ISDIR_START	15
#define AFTER_DIR	" "
#define SIZE_COL	15	/* column with size of file */
#define DATESTART	26	/* column with file date */
#endif
#endif
#if 0
#define FILECHARS	"[^ \n]*"	/* legal chars in filename */
#else
#define FILECHARS	".*"
#endif

#define DIRLINEPAT	"^     Directory of "

when_loading()		/* disable self-inserting chars */
{
	fix_key_table(reg_tab, (short) normal_character, dired_tab, -1);
}

dired_help() on dired_tab['h'], dired_tab['?']
{
	help_on_command("dired-mode");
}

dired_examine() on dired_tab[' '], dired_tab['\n'],
		   dired_tab['e'], dired_tab['\r'], dired_tab[GREYENTER]
{
	char fname[FNAMELEN];

	dired_getname(fname);
	do_dired_examine(fname);
	iter = 0;
}

dired_parent() on dired_tab['^'], dired_tab['.']
{				/* view parent directory */
	char fname[FNAMELEN];

	to_begin_line();
	dired_try_getdir(fname, 1);
	strcat(fname, "..");
	do_dired_examine(fname);
	iter = 0;
}

do_dired_examine(fname)
char *fname;
{
	char file[FNAMELEN], *p;

	strcpy(file, fname);
	absolute(file);
	if (is_directory(file)) {
		if (dired_one(file))
			return;
		strcpy(file, fname);
		p = get_tail(file, 0);	/* for .., go to right place */
		if (p > file && strcmp(p, "..") == 0) {
			strcpy(p - 1, AFTER_DIR);
			col_search(get_tail(file, 0), NAME_START);
			move_to_column(goal_column);
			build_first = 1;
		}
	} else			/* regular file */
		find_it(file, has_arg ? !strip_returns.default
				: strip_returns.default);
}

dired_examine_in_window() on dired_tab['1'], dired_tab['2'],
				dired_tab['5'], dired_tab['z']
{		/* do appropriate window cmd, then examine */
	int olditer = iter, oldarg = has_arg;

	has_arg = iter = 0;
	do_command(cx_tab[key]);
	iter = olditer, has_arg = oldarg;
	ungot_key = 'e';
}

dired_examine_in_next_window() on dired_tab['o']
{
	char fname[FNAMELEN];

	dired_getname(fname);
	next_user_window(1);
	do_dired_examine(fname);
	iter = 0;
}

dired_refresh() on dired_tab['r']
{
	char msg[FNAMELEN];
	int yes;

	if (modified) {
		sprintf(msg, "Relist %s, discarding markings? ", bufname);
		yes = ask_yn("Discard markings", msg, "Yes", "No", 1);
		check_abort();
		if (!yes)
			return;
	}
	save_var point, window_start;
	dired_one(bufname);
	restore_vars();
	move_to_column(goal_column);
	fix_window_start();
}

dired_sort() on dired_tab['s']
{
	char *p, msg[200];
	int rev, sort;

	rev = isupper(dired_sorts_files);
	switch (sort = toupper(dired_sorts_files)) {
		case 'N': case 'D': case 'S':  case 'E': 
			break;
		case 0: case 'U':
			sort = 'U', rev = 0; break;
		default:
			sort = 'N';
	}
	for (;;) {
		strcpy(msg, "Name, Ext, Size, Date, Unsorted,");
		p = index(msg, sort);
		p = index(p, ',');
		msg[strlen(msg) - 1] = 0;
		sayput("Sort by %.*s (now)%s; Reverse; %sGroup dirs"
		       " [NESDURG] ", p - msg, msg, p,
		       dired_groups_dirs ? "un" : "");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'N':
			case 'D':
			case 'S':
			case 'E': sort = toupper(key); break;
			case 'U': say("Future listings will be unsorted.");
				  sort = 0; break;
			case 'R': rev = !rev; break;
			case '+': rev = 0; break;
			case '-': rev = 1; break;
			case 'G': dired_groups_dirs = !dired_groups_dirs;
				  break;
			default:  if (IS_ENTER_KEY(key))
					  break;
				  else if (root_keys[key] == (short) help
					   || key == 'h' || key == '?')
					  help_on_command("dired-sort");
				  continue;
		}
		break;
	}
	dired_sorts_files = rev ? toupper(sort) : tolower(sort);
	if (sort) {
		sprintf(msg, "%s%s, %s", rev ? "Reverse " : "",
			(sort == 'N') ? "Name" : (sort == 'E') ? "Extension"
			: (sort == 'S') ? "Size" : "Date",
			dired_groups_dirs ? "grouping directories at top"
					  : "not grouping directories");
		noteput("Sorting...");
		sort_for_dired();
		say("Sorted by %s.", msg);
	} else
		say("Future listings will be unsorted.");
	maybe_refresh();
}

dired_repl(newchar)
{
	to_begin_line();
	if (parse_string(1, DIRLINEPAT))
		nl_forward();
	if (index("DCM ", curchar()) && character(point + 1) == ' ')
		replace(point, newchar);
	save_var iter = 1;		// call us to do each repetition
	dired_down_line();
}

dired_mark_for_delete() on dired_tab['d']
{
	dired_repl('D');
}

dired_mark_for_copy() on dired_tab['c']
{
	dired_repl('C');
}

dired_mark_for_move() on dired_tab['m']
{
	dired_repl('M');
}

dired_unmark() on dired_tab['u']
{
	dired_repl(' ');
}

dired_do_delete(name, to)	/* return 1 if error */
char *name, *to;
{
	to = to;
	noteput("Deleting %s.", name);
	if (dired_isdir() ? rmdir(name) : delete_file(name))
		return 1;
	save_var modified;	// leave modify bit alone
	to_begin_line();	// but delete line
	delete(point - 1, give_end_line());
	return 0;
}

dired_do_copy(name, to)	/* return 1 if error */
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	noteput("Copying %s to %s.", name, dest);
	if (!strcmp(name, dest) || warn_existing_file(dest)
	    || copyfile(name, dest))
		return 1;	// disallow copying file to itself
	save_var modified;	// leave modify bit alone
	dired_unmark();
	nl_reverse();
	return 0;
}

dired_do_move(name, to)	/* return 1 if error */
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	noteput("Moving %s to %s.", name, dest);
	if (rename_file(name, dest))
		return 1;
	save_var modified;	// leave modify bit alone
	to_begin_line();	// but delete line
	delete(point - 1, give_end_line());
	return 0;
}

make_copy_name(src, dest, res)	// if dest is a dir, append src to it
char *src, *dest, *res;		// else copy dest
{
	strcpy(res, dest);
	if (is_directory(dest)) {
		dest = res + strlen(res);
		if (*res && !is_path_separator(dest[-1]))
			*dest++ = path_sep;
		strcpy(dest, get_tail(src, 0));
	}
}

dired_execute() on dired_tab['q'], dired_tab['x']
{
	int del, copy, move;

	iter = 0;
	del = dired_do_no_param("Delete", "\nD ", "deleted", dired_do_delete);
	copy = dired_do_param("Copy", "\nC ", "copied", dired_do_copy);
	move = dired_do_param("Move", "\nM ", "moved", dired_do_move);
	if (del + copy + move == 0)
		error("No files marked");
	modified = 0;
	if (move || copy)	// could be rename or similar, make new listing
		dired_refresh();
	fix_window_start();
}

dired_do_no_param(op, tag, past, func)
char *op, *tag, *past;
int (*func)();
{
	int count = dired_ask(op, tag, "? ", NULL);

	if (count > 0)
		dired_do(past, tag, count, func, NULL);
	return count;
}

dired_do_param(op, tag, past, func)
char *op, *tag, *past;
int (*func)();
{
	char to[FNAMELEN];
	int count = dired_ask(op, tag, " to: ", to);

	if (count > 1)		// can only copy/move mult to a dir
		switch (check_file(to, NULL)) {
		case 0:		if (mkdir(to)) {
					file_error(errno, to, "make dir");
					quick_abort();
				}
		case CHECK_DIR:	break;
		default:	error("%s is not a directory.", to);
		}
	if (count > 0)
		dired_do(past, tag, count, func, to);
	return count;
}

dired_ask(op, tag, end, file)	// make list of files, confirm/get param if any
char *op, *tag, *end, *file;	// return count or 0 if none were marked
{
	int list = tmp_buf(), count = 0;
	char msg[FNAMELEN], name[FNAMELEN];

	save_var point = 0;
	save_var abort_searching = ABORT_ERROR;
	while (search(1, tag) == 1) {
		dired_getname(name);
		count++;
		if (count > 1)
			sprintf(msg, "%s these %d files%s", op, count, end);
		else
			sprintf(msg, "%s this file%s", op, end);
		if (count > 1 && !IS_GUI)
			note(msg);
		buf_printf(list, "%s\n", name);
	}
	restore_vars();
	if (!count)
		return 0;
	else if (count == 1 && (!IS_GUI || file))
		sprintf(msg, "%s file %s%s", op, name, end);
	else {
		if (!IS_GUI || file)
			say(msg);
		sprintf(name, " Files to %s ", op);
		save_var _view_title = (IS_GUI && !file) ? msg : name;
		buf_delete_final_newlines(list);
		view_buf(list, 0);
	}
	save_var prompt_with_buffer_directory = PBD_INSERT;
	if (is_gui && key == abort_key)
		file = 0, *name = 'N';		// User clicked Cancel.
	else if (file)
		get_file_dir(file, msg);
	else if (IS_GUI) {	// Already prompted so don't ask again.
		*name = toupper(key) == 'Y' || IS_ENTER_KEY(key)
				|| toupper(key) == 'O' ? 'Y' : 'N';
		ungot_key = -1;
	} else
		get_string(name, msg);
	buf_delete(list);
	if (file ? *file : (toupper(*name) == 'Y'))
		return count;
	error("No action taken.");
}

dired_do(op, tag, count, func, to)
char *op, *tag, *to;
int (*func)();
{
	spot final;
	int bad = 0;
	char name[FNAMELEN];

	final = alloc_spot();
	point = 0;
	save_var abort_searching = 0;
	while (search(1, tag)) {
		if (user_abort) {
			if (!bad)
				*final = point;
			break;
		}
		dired_getname(name);
		if ((*func)(name, to))
			if (!bad++)		/* error occurred */
				*final = point;	/* 1st error */
	}
	point = *final;
	free_spot(final);
	move_to_column(goal_column);
	if (bad)
		error("%d of %d file%s not %s", bad, count,
			count > 1 ? "s" : "", op);
	say("%d file%s %s", count, count > 1 ? "s" : "", op);
}

dired_del_line()	/* delete current line, leave modify bit alone */
{
	save_var modified;	// leave modify bit alone
	to_begin_line();	// but delete line
	delete(point, give_end_line() + 1);
}

dired_run_viewer() on dired_tab['v']
{
	char fname[FNAMELEN], dirname[FNAMELEN];

	dired_getname(fname);
	strcpy(dirname, fname);
	*get_tail(dirname, 1) = 0;
	if ( run_viewer(fname, "open", dirname) == 0)
		say("Could not run viewer for \"%s\"", fname);
}

dired_goto_dir() on dired_tab['g']
{
	char fname[FNAMELEN];

	iter = 0;
	dired_getname(fname);
	if (chdir(fname) && (dired_try_getdir(fname, 1), chdir(fname)))
		file_error(errno, fname, "unknown error");
	else {
		getcd(fname);
		say("Current directory: %s", fname);
	}
}

dired_down_line() on dired_tab['n']
{
	down_line();
	to_begin_line();
	if (parse_string(1, DIRLINEPAT))
		nl_forward();
	move_to_column(goal_column);
}

dired_up_line() on dired_tab['p']
{
	up_line();
	to_begin_line();
	if (parse_string(1, DIRLINEPAT))
		nl_reverse();
	move_to_column(goal_column);
}

dired_getdir(fullname)	// grab directory, add / to end, for completion
char *fullname;		// set fullname to "" if error
{
	return dired_try_getdir(fullname, 0);
}

dired_try_getdir(fullname, showerr)	// grab directory, add / to end
char *fullname;		// if we can't find, set fullname "" if !showerr
{			// return nonzero if ok
	int i;
	char *p;

	*fullname = 0;
	save_var point;
	to_end_line();
	if (!re_search(-1, DIRLINEPAT))	// find prev "Directory of "
		if (showerr)		// bad dired buf
			error("Can't determine directory of files");
		else
			return 0;
	point = matchstart;
	re_search(1, " *");		// skip initial spaces
	parse_string(1, FILECHARS, fullname);
	p = fullname + (i = strlen(fullname));
	if (i > 0 && !is_path_separator(fullname[i - 1]))
		p[0] = path_sep, p[1] = 0; // add final / if none
	return 1;
}

dired_getname(fullname)
char *fullname;
{
	char *p;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, NAME_START);		/* the file name */
	to_begin_line();
	if (parse_string(1, DIRLINEPAT))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		error("Can't read file name");
#ifdef UNIX
	p += strlen(p) - 1;
	if (*p == '/' || *p == '*')
		*p = 0;
#endif
}

dired_isdir()
{
	char buf[FNAMELEN];

	dired_getindex(buf, ISDIR_START);
	return !strncmp(buf, DIRMARK, strlen(DIRMARK));
}

dired_getindex(s,i)	/* put current line in s, starting at column i */
char *s;
{
	save_var point;
	move_to_column(i);
	re_search(1, " *");		/* skip initial spaces */
	parse_string(1, FILECHARS, s);
}

is_dired_buf()		/* is the current buffer a dired buffer? */
{
	return mode_keys == dired_tab;
}


command dired() on cx_tab['d']
{
	char dir[FNAMELEN];

	tiled_only();
	get_file_dir(dir, "Edit directory: ");
	return dired_one(dir);
}

char _dired_mode_name[] = "Dired";

dired_mode()
{
	mode_keys = dired_tab;
	major_mode = _dired_mode_name;
	mouse_dbl_selects = 1;
	goal_column = NAME_START;
	discardable_buffer = 1;
	make_mode();
}

dired_one(dir)	/* dir is absolute pathname of directory or pattern */
char *dir;	/* return 0 if ok */
{
	char buf[FNAMELEN], *to;
	int uniq = 1;

	if (*dir)
		strcpy(buf, dir);
	else if (prompt_with_buffer_directory) {
		get_buffer_directory(buf);	// this file's directory
		dir = buf;
	} else {
		dir = ".";	/* "" means current directory */
		getcd(buf);
	}
	dired_standardize(buf);		/* fix appropriately for this o/s */
	to = buf + strlen(buf);
	while (exist(buf)) {	/* make it unique */
		bufname = buf;
		if (is_dired_buf())	/* reuse existing dired buffer */
			break;
		sprintf(to, "<%d>", uniq++);
	}
	to_buffer(buf);
	zap(buf);
	dired_mode();
	noteput("Building directory listing...");
	if (make_dired(dir)) {
		say("No matches.");
		return 1;
	}
	say("");
	return 0;
}

make_dired(dir)
char *dir;
{
	modified = 0;
	save_var abort_file_matching = ABORT_JUMP;
	if (do_dired(dir))
		return 1;
	switching_to_buffer();
	modified = 0;
	sort_for_dired();
	return 0;
}

sort_for_dired()
{
	int start, done = 0;

	save_var modified;
	window_start = point = 0;
	save_var abort_searching = 0;
	while (!done) {
		do {			/* skip past . and .. */
			nl_forward();	/* exclude 1st line */
			move_to_column(NAME_START);
		} while (parse_string(1, "<.><.>?[/\\]?($| )"));
		to_begin_line();
		start = point;
		if (re_search(1, DIRLINEPAT))
			to_begin_line();
		else
			done = 1;
		dired_sorter(start, point);
	}
	point = 0;
	dired_down_line();		/* position on 2nd line */
	window_start = 0;
}

dired_sorter(start, end)
{
	int tmp, next;

	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	if (dired_groups_dirs) {
		save_var point = 0;
		tmp = tmp_buf();
		while (re_search(1, DIRSEARCH) && !user_abort) {
			to_begin_line();
			next = give_end_line() + 1;
			buf_xfer(tmp, point, next);
			delete(point, next);
		}
		dired_group_sorter(start, point);
		point = 0;
		grab_buffer(tmp);
		buf_delete(tmp);
		dired_group_sorter(mark, point);
	} else
		dired_group_sorter(start, end);
}

dired_group_sorter(start, end)
{
	int on_col = 0, stripwid = 0;

	if (!dired_sorts_files || user_abort)
		return;
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	save_var point;
	save_var abort_searching = ABORT_ERROR;
	switch (toupper(dired_sorts_files)) {
		case 'D': stripwid = dired_sort_by_date(); break;
		case 'E': stripwid = dired_sort_by_exten(); break;
		case 'S': on_col = SIZE_COL; break;
		case 'N':
		default: on_col = NAME_START; break;
	}
	do_sort_region(0, size(), on_col, isupper(dired_sorts_files));
	if (stripwid)
		strip_columns(0, stripwid);
}

dired_sort_by_date()	/* insert stuff before each line, return length */
{
	char line[FNAMELEN];
	int end, hour;

	if (size() - narrow_end - narrow_start > 10000)
		noteput("Sorting by date...");
	point = 0;
	do {
		end = give_end_line();
		if (point == end)
			break;
		grab(point, end, line);
		hour = numtoi(line + DATESTART + 12);
		if (hour == 12)
			hour = 0;
		if (line[DATESTART + 17] == 'p')
			hour += 12;
		bprintf("%.2s%.5s%2d%.2s", line + DATESTART + 8,
			line + DATESTART + 2, hour, line + DATESTART + 15);
		point = end;
	} while (nl_forward());
	return 11;
}

dired_sort_by_exten()	/* insert stuff before each line, return length */
{
	char line[FNAMELEN];

	if (size() - narrow_end - narrow_start > 10000)
		noteput("Sorting by extension...");
	point = 0;
	do {
		if (point == give_end_line())
			break;
		dired_getindex(line, NAME_START);
		bprintf("%-30.30s", get_extension(line));
	} while (nl_forward());
	return 30;
}

strip_columns(scol, ecol)
{
	point = 0;
	do {
		delete(column_to_pos(scol), column_to_pos(ecol));
	} while (nl_forward());
}

when_loading()
{
	set_case_indirect(dired_tab);
}
